# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import annotations

import decimal as dec
import posixpath
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from ase.atoms import Atoms
from ase.data import atomic_masses, atomic_numbers

from pyiron_lammps.units import UnitConverter

try:
    from ase.calculators.lammps import Prism
except ImportError:
    try:
        from ase.calculators.lammpsrun import Prism
    except ImportError:
        from ase.calculators.lammpsrun import prism as Prism

__author__ = "Joerg Neugebauer, Sudarsan Surendralal, Yury Lysogorskiy, Jan Janssen, Markus Tautschnig"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class UnfoldingPrism(Prism):
    """
    Create a lammps-style triclinic prism object from a cell

    The main purpose of the prism-object is to create suitable
    string representations of prism limits and atom positions
    within the prism.
    When creating the object, the digits parameter (default set to 10)
    specify the precision to use.
    lammps is picky about stuff being within semi-open intervals,
    e.g. for atom positions (when using create_atom in the in-file),
    x must be within [xlo, xhi).

    Args:
        cell:
        pbc:
        digits:
    """

    def __init__(
        self,
        cell: np.ndarray,
        pbc: Union[bool, tuple[bool, bool, bool]] = (True, True, True),
        digits: int = 10,
    ):
        # Temporary fix. Since the arguments for the constructor have changed, try to see if it is compatible with
        # the latest ase. If not, revert to the old __init__ parameters.
        if isinstance(pbc, bool):
            pbc = (pbc, pbc, pbc)
        try:
            super(UnfoldingPrism, self).__init__(
                cell, pbc=np.array(pbc), tolerance=float("1e-{}".format(digits))
            )
        except TypeError:
            super(UnfoldingPrism, self).__init__(cell, pbc=np.array(pbc), digits=digits)
        a, b, c = cell
        an, bn, cn = [np.linalg.norm(v) for v in cell]

        alpha = np.arccos(np.dot(b, c) / (bn * cn))
        beta = np.arccos(np.dot(a, c) / (an * cn))
        gamma = np.arccos(np.dot(a, b) / (an * bn))

        xhi = an
        xyp = np.cos(gamma) * bn
        yhi = np.sin(gamma) * bn
        xzp = np.cos(beta) * cn
        yzp = (bn * cn * np.cos(alpha) - xyp * xzp) / yhi
        zhi = np.sqrt(cn**2 - xzp**2 - yzp**2)

        # Set precision
        self.car_prec = dec.Decimal("10.0") ** int(
            np.floor(np.log10(max((xhi, yhi, zhi)))) - digits
        )
        self.dir_prec = dec.Decimal("10.0") ** (-digits)
        self.acc = float(self.car_prec)
        self.eps = np.finfo(xhi).eps

        # For rotating positions from ase to lammps
        apre = np.array(((xhi, 0, 0), (xyp, yhi, 0), (xzp, yzp, zhi)))
        # np.linalg.inv(cell) ?= np.array([np.cross(b, c), np.cross(c, a), np.cross(a, b)]).T / np.linalg.det(cell)
        self.R = np.dot(np.linalg.inv(cell), apre)

        def fold(
            vec: np.ndarray, pvec: np.ndarray, i: int
        ) -> Tuple[List[float], float]:
            p = pvec[i]
            x = vec[i] + 0.5 * p
            n = (np.mod(x, p) - x) / p
            return [float(self.f2qdec(vec_a)) for vec_a in (vec + n * pvec)], n

        apre[1, :], n1 = fold(apre[1, :], apre[0, :], 0)
        if np.abs(apre[1, 0] / apre[0, 0]) > 0.5:
            apre[1, 0] -= np.sign(n1) * apre[0, 0]
            n1 -= np.sign(n1)

        apre[2, :], n2 = fold(apre[2, :], apre[1, :], 1)
        if np.abs(apre[2, 1] / apre[1, 1]) > 0.5:
            apre[2, 1] -= np.sign(n2) * apre[1, 1]
            n2 -= np.sign(n2)

        apre[2, :], n3 = fold(apre[2, :], apre[0, :], 0)
        if np.abs(apre[2, 0] / apre[0, 0]) > 0.5:
            apre[2, 0] -= np.sign(n3) * apre[0, 0]
            n3 -= np.sign(n3)
        self.ns = [n1, n2, n3]

        d_a = apre[0, 0] / 2 - apre[1, 0]
        if np.abs(d_a) < self.acc:
            if d_a < 0:
                print("debug: apply shift")
                apre[1, 0] += 2 * d_a
                apre[2, 0] += 2 * d_a

        self.A = apre

        if self.is_skewed() and (not (pbc[0] and pbc[1] and pbc[2])):
            warnings.warn(
                "Skewed lammps cells should have PBC == True in all directions!"
            )

    def unfold_cell(self, cell: np.ndarray) -> np.ndarray:
        """
        Unfold LAMMPS cell to original

        Let C be the pyiron_atomistics cell and A be the Lammps cell, then define (in init) the rotation matrix between them as
            R := C^inv.A
        And recall that rotation matrices have the property
            R^T == R^inv
        Then left multiply the definition of R by C, and right multiply by R.T to get
            C.R.R^T = C.C^inv.A.R^T
        Then
            C = A.R^T

        After that, account for the folding process.

        Args:
            cell: LAMMPS cell,

        Returns:
            unfolded cell
        """
        # Rotation
        ucell = np.dot(cell, self.R.T)
        # Folding
        a = ucell[0]
        bp = ucell[1]
        cpp = ucell[2]
        n1, n2, n3 = self.ns
        b = bp - n1 * a
        c = cpp - n2 * bp - n3 * a
        return np.array([a, b, c])

    def pos_to_lammps(self, position: np.ndarray) -> Tuple[float, float, float]:
        """
        Rotate an ase-cell position to the lammps cell orientation

        Args:
            position:

        Returns:
            tuple of float.
        """
        return tuple([x for x in np.dot(position, self.R)])

    def f2qdec(self, f: float) -> dec.Decimal:
        return dec.Decimal(str(f)).quantize(self.car_prec, dec.ROUND_DOWN)

    def f2s(self, f: float) -> str:
        return str(dec.Decimal(str(f)).quantize(self.car_prec, dec.ROUND_HALF_EVEN))

    def get_lammps_prism_str(self) -> Tuple[str, ...]:
        """Return a tuple of strings"""
        p = self.get_lammps_prism()
        return tuple([self.f2s(x) for x in p])


class LammpsStructure:
    """

    Args:
        input_file_name:
    """

    def __init__(
        self,
        bond_dict: Optional[Dict] = None,
        units: str = "metal",
        atom_type: str = "atomic",
    ):
        self._string_input: str = ""
        self._structure: Optional[Atoms] = None
        self._potential: Optional[Any] = None
        self._el_eam_lst: List[str] = []
        self.atom_type: Optional[str] = None
        self.cutoff_radius: Optional[float] = None
        self.digits: int = 10
        self._bond_dict: Optional[Dict] = bond_dict
        self._force_skewed: bool = False
        self._units: str = units
        self._atom_type: str = atom_type

    @property
    def potential(self) -> Any:
        return self._potential

    @potential.setter
    def potential(self, val: Any):
        self._potential = val

    @property
    def structure(self) -> Optional[Atoms]:
        """

        Returns:

        """
        return self._structure

    @structure.setter
    def structure(self, structure: Atoms):
        """

        Args:
            structure:

        Returns:

        """
        self._structure = structure
        if self._atom_type == "charge":
            input_str = self.structure_charge()
        else:  # self.atom_type == 'atomic'
            input_str = self.structure_atomic()
        self._string_input = input_str + self._get_velocities_input_string()

    def _get_velocities_input_string(self) -> str:
        input_str = ""
        if self._structure is not None and not np.all(
            np.isclose(
                self._structure.get_velocities(),
                np.array([[0.0, 0.0, 0.0]] * len(self._structure)),
            )
        ):
            uc = UnitConverter(self._units)
            self._structure.set_velocities(
                self._structure.get_velocities() * uc.pyiron_to_lammps("velocity")
            )
            vels = self.rotate_velocities(self._structure)
            input_str += "Velocities\n\n"
            if len(self._structure.positions[0]) == 3:
                format_str = "{0:d} {1:f} {2:f} {3:f}\n"
                for id_atom, (x, y, z) in enumerate(vels, start=1):
                    input_str += format_str.format(id_atom, x, y, z)
            elif len(self._structure.positions[0]) == 2:
                format_str = "{0:d} {1:f} {2:f}\n"
                for id_atom, (x, y) in enumerate(vels, start=1):
                    input_str += format_str.format(id_atom, x, y)
        return input_str

    @property
    def el_eam_lst(self) -> List[str]:
        """

        Returns:

        """
        return self._el_eam_lst

    @el_eam_lst.setter
    def el_eam_lst(self, el_eam_lst: List[str]):
        """

        Args:
            el_eam_lst:

        Returns:

        """
        self._el_eam_lst = el_eam_lst

    @staticmethod
    def get_lammps_id_dict(el_eam_lst: List[str]) -> Dict[str, int]:
        if len(el_eam_lst) == 0:
            raise ValueError("el_eam_list is empty. Can not determine order of species")
        return {el: idx + 1 for idx, el in enumerate(el_eam_lst)}

    @staticmethod
    def lammps_header(
        structure: Atoms,
        cell_dimensions: str,
        species_lammps_id_dict: Dict[str, int],
        nbonds: Optional[int] = None,
        nangles: Optional[int] = None,
        nbond_types: Optional[int] = None,
        nangle_types: Optional[int] = None,
    ) -> str:
        atomtypes = (
            "Start File for LAMMPS \n"
            + "{0:d} atoms".format(len(structure))
            + " \n"
            + "{0} atom types".format(len(species_lammps_id_dict.keys()))
            + " \n"
        )  # '{0} atom types'.format(structure.get_number_of_species()) + ' \n'
        if nbonds is not None:
            atomtypes += "{0:d} bonds\n".format(nbonds)
        if nangles is not None:
            atomtypes += "{0:d} angles\n".format(nangles)
        if nbond_types is not None:
            atomtypes += "{0:d} bond types\n".format(nbond_types)
        if nangle_types is not None:
            atomtypes += "{0:d} angle types\n".format(nangle_types)

        masses = "Masses\n\n"
        for el, idx in species_lammps_id_dict.items():
            mass = atomic_masses[atomic_numbers[el]]
            masses += "{0:3d} {1:f}  # ({2}) \n".format(idx, mass, el)

        return atomtypes + "\n" + cell_dimensions + "\n" + masses + "\n"

    def simulation_cell(self) -> str:
        """

        Returns:

        """
        if self._structure is None:
            raise ValueError("Structure not set")
        self.prism = UnfoldingPrism(self._structure.cell, digits=15)
        xhi, yhi, zhi, xy, xz, yz = self.prism.get_lammps_prism_str()
        # Please, be carefull and not round xhi, yhi,..., otherwise you will get too skew cell from LAMMPS.
        # These values are already checked in UnfoldingPrism to fullfill LAMMPS skewness criteria
        simulation_cell = (
            "0. {} xlo xhi\n".format(xhi)
            + "0. {} ylo yhi\n".format(yhi)
            + "0. {} zlo zhi\n".format(zhi)
        )

        if is_skewed(self.structure) or self._force_skewed:
            simulation_cell += "{0} {1} {2} xy xz yz\n".format(xy, xz, yz)

        return simulation_cell

    def structure_atomic(self) -> str:
        """
        Write routine to create atom structure static file that can be loaded by LAMMPS

        Returns:

        """
        if self._structure is None:
            raise ValueError("Structure not set")
        species_lammps_id_dict = self.get_lammps_id_dict(self.el_eam_lst)
        atoms = "Atoms\n\n"
        coords = self.rotate_positions(self._structure)

        el_lst = self._structure.get_chemical_symbols()
        for id_atom, (el, coord) in enumerate(zip(el_lst, coords)):
            dim = len(self._structure.positions[0])
            c = np.zeros(3)
            c[:dim] = coord
            atoms += (
                "{0:d} {1:d} {2:.15f} {3:.15f} {4:.15f}".format(
                    id_atom + 1, species_lammps_id_dict[el], c[0], c[1], c[2]
                )
                + "\n"
            )
        return (
            self.lammps_header(
                structure=self.structure,
                cell_dimensions=self.simulation_cell(),
                species_lammps_id_dict=species_lammps_id_dict,
            )
            + atoms
            + "\n"
        )

    def structure_charge(self):
        """
        Create atom structure including the atom charges.

        By convention the LAMMPS atom type numbers are chose alphabetically for the chemical species.

        Returns: LAMMPS readable structure.

        """
        species_lammps_id_dict = self.get_lammps_id_dict(self.el_eam_lst)
        atoms = "Atoms\n\n"
        coords = self.rotate_positions(self._structure)
        el_charge_lst = self._structure.get_initial_charges()
        el_lst = self._structure.get_chemical_symbols()
        for id_atom, (el, coord) in enumerate(zip(el_lst, coords)):
            dim = len(self._structure.positions[0])
            c = np.zeros(3)
            c[:dim] = coord
            atoms += (
                "{0:d} {1:d} {2:f} {3:.15f} {4:.15f} {5:.15f}".format(
                    id_atom + 1,
                    species_lammps_id_dict[el],
                    el_charge_lst[id_atom],
                    c[0],
                    c[1],
                    c[2],
                )
                + "\n"
            )
        return (
            self.lammps_header(
                structure=self.structure,
                cell_dimensions=self.simulation_cell(),
                species_lammps_id_dict=species_lammps_id_dict,
            )
            + atoms
            + "\n"
        )

    def rotate_positions(self, structure: Atoms) -> List[Tuple[float, float, float]]:
        """
        Rotate all atomic positions in given structure according to new Prism cell

        Args:
            structure: Atoms-like object. Should has .positions attribute

        Returns:
            (list): List of rotated coordinates
        """
        if self._structure is None:
            raise ValueError("Structure not set")
        prism = UnfoldingPrism(self._structure.cell)
        coords = [prism.pos_to_lammps(position) for position in structure.positions]
        return coords

    def rotate_velocities(self, structure: Atoms) -> List[Tuple[float, float, float]]:
        """
        Rotate all atomic velocities in given structure according to new Prism cell

        Args:
            structure: Atoms-like object. Should have .velocities attribute.

        Returns:
            (list): List of rotated velocities
        """
        if self._structure is None:
            raise ValueError("Structure not set")
        prism = UnfoldingPrism(self._structure.cell)
        vels = [prism.pos_to_lammps(vel) for vel in structure.get_velocities()]
        return vels

    def write_file(self, file_name: str, cwd: Optional[str] = None):
        """
        Write GenericParameters to input file

        Args:
            file_name (str): name of the file, either absolute (then cwd must be None) or relative
            cwd (str): path name (default: None)
        """
        if cwd is not None:
            file_name = posixpath.join(cwd, file_name)

        with open(file_name, "w") as f:
            for line in self._string_input:
                f.write(line)


def is_skewed(structure: Atoms, tolerance: float = 1.0e-8) -> bool:
    """
    Check whether the simulation box is skewed/sheared. The algorithm compares the box volume
    and the product of the box length in each direction. If these numbers do not match, the box
    is considered to be skewed and the function returns True

    Args:
        tolerance (float): Relative tolerance above which the structure is considered as skewed

    Returns:
        (bool): Whether the box is skewed or not.
    """
    volume = structure.get_volume()
    prod = np.linalg.norm(structure.cell, axis=-1).prod()
    if volume > 0:
        if abs(volume - prod) / volume < tolerance:
            return False
    return True


def write_lammps_datafile(
    structure: Atoms,
    potential_elements: Union[np.ndarray, list],
    bond_dict: Optional[Dict] = None,
    units: str = "metal",
    file_name: str = "lammps.data",
    working_directory: Optional[str] = None,
    atom_type: str = "atomic",
) -> None:
    lammps_str = LammpsStructure(bond_dict=bond_dict, units=units, atom_type=atom_type)
    lammps_str.el_eam_lst = potential_elements
    lammps_str.structure = structure
    lammps_str.write_file(file_name=file_name, cwd=working_directory)


def structure_to_lammps(structure: Atoms) -> Atoms:
    """
    Converts a structure to the Lammps coordinate frame

    Args:
        structure (pyiron.atomistics.structure.atoms.Atoms): Structure to convert.

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: Structure with the LAMMPS coordinate frame.
    """
    prism = UnfoldingPrism(structure.cell)
    lammps_structure = structure.copy()
    lammps_structure.set_cell(prism.A)
    lammps_structure.positions = np.matmul(structure.positions, prism.R)
    if not np.all(
        np.isclose(
            structure.get_velocities(), np.array([[0.0, 0.0, 0.0]] * len(structure))
        )
    ):
        lammps_structure.set_velocities(np.matmul(structure.get_velocities(), prism.R))
    return lammps_structure



import numpy as np
from ase.atoms import Atoms


def _get_fixed_atom_boolean_vector(structure: Atoms):
    fixed_atom_vector = np.array([[False, False, False]] * len(structure))
    for c in structure.constraints:
        c_dict = c.todict()
        if c_dict["name"] == "FixAtoms":
            fixed_atom_vector[c_dict["kwargs"]["indices"]] = [True, True, True]
        elif c_dict["name"] == "FixedPlane":
            if all(np.isin(c_dict["kwargs"]["direction"], [0, 1, 1 / np.sqrt(2)])):
                if "indices" in c_dict["kwargs"]:
                    fixed_atom_vector[c_dict["kwargs"]["indices"]] = np.array(
                        c_dict["kwargs"]["direction"]
                    ).astype(bool)
                elif "a" in c_dict["kwargs"]:
                    fixed_atom_vector[c_dict["kwargs"]["a"]] = np.array(
                        c_dict["kwargs"]["direction"]
                    ).astype(bool)
            else:
                raise ValueError(
                    "Currently the directions are limited to [1, 0, 0], [1, 1, 0], [1, 1, 1] and its permutations."
                )
        else:
            raise ValueError("Only FixAtoms and FixedPlane are currently supported. ")
    return fixed_atom_vector


def set_selective_dynamics(structure: Atoms, calc_md: bool = False):
    control_dict = {}
    if len(structure.constraints) > 0:
        sel_dyn = _get_fixed_atom_boolean_vector(structure=structure)
        # Enter loop only if constraints present
        if len(np.argwhere(np.any(sel_dyn, axis=1)).flatten()) != 0:
            all_indices = np.arange(len(structure), dtype=int)
            constraint_xyz = np.argwhere(np.all(sel_dyn, axis=1)).flatten()
            not_constrained_xyz = np.setdiff1d(all_indices, constraint_xyz)
            # LAMMPS starts counting from 1
            constraint_xyz += 1
            ind_x = np.argwhere(sel_dyn[not_constrained_xyz, 0]).flatten()
            ind_y = np.argwhere(sel_dyn[not_constrained_xyz, 1]).flatten()
            ind_z = np.argwhere(sel_dyn[not_constrained_xyz, 2]).flatten()
            constraint_xy = not_constrained_xyz[np.intersect1d(ind_x, ind_y)] + 1
            constraint_yz = not_constrained_xyz[np.intersect1d(ind_y, ind_z)] + 1
            constraint_zx = not_constrained_xyz[np.intersect1d(ind_z, ind_x)] + 1
            constraint_x = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_x, ind_y), ind_z)] + 1
            )
            constraint_y = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_y, ind_z), ind_x)] + 1
            )
            constraint_z = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_z, ind_x), ind_y)] + 1
            )
            control_dict = {}
            if len(constraint_xyz) > 0:
                control_dict["group constraintxyz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_xyz]
                )
                control_dict["fix constraintxyz"] = "constraintxyz setforce 0.0 0.0 0.0"
                if calc_md:
                    control_dict["velocity constraintxyz"] = "set 0.0 0.0 0.0"
            if len(constraint_xy) > 0:
                control_dict["group constraintxy"] = "id " + " ".join(
                    [str(ind) for ind in constraint_xy]
                )
                control_dict["fix constraintxy"] = "constraintxy setforce 0.0 0.0 NULL"
                if calc_md:
                    control_dict["velocity constraintxy"] = "set 0.0 0.0 NULL"
            if len(constraint_yz) > 0:
                control_dict["group constraintyz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_yz]
                )
                control_dict["fix constraintyz"] = "constraintyz setforce NULL 0.0 0.0"
                if calc_md:
                    control_dict["velocity constraintyz"] = "set NULL 0.0 0.0"
            if len(constraint_zx) > 0:
                control_dict["group constraintxz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_zx]
                )
                control_dict["fix constraintxz"] = "constraintxz setforce 0.0 NULL 0.0"
                if calc_md:
                    control_dict["velocity constraintxz"] = "set 0.0 NULL 0.0"
            if len(constraint_x) > 0:
                control_dict["group constraintx"] = "id " + " ".join(
                    [str(ind) for ind in constraint_x]
                )
                control_dict["fix constraintx"] = "constraintx setforce 0.0 NULL NULL"
                if calc_md:
                    control_dict["velocity constraintx"] = "set 0.0 NULL NULL"
            if len(constraint_y) > 0:
                control_dict["group constrainty"] = "id " + " ".join(
                    [str(ind) for ind in constraint_y]
                )
                control_dict["fix constrainty"] = "constrainty setforce NULL 0.0 NULL"
                if calc_md:
                    control_dict["velocity constrainty"] = "set NULL 0.0 NULL"
            if len(constraint_z) > 0:
                control_dict["group constraintz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_z]
                )
                control_dict["fix constraintz"] = "constraintz setforce NULL NULL 0.0"
                if calc_md:
                    control_dict["velocity constraintz"] = "set NULL NULL 0.0"
    return control_dict



import warnings

import numpy as np
from ase.atoms import Atoms

from pyiron_lammps.structure import UnfoldingPrism
from pyiron_lammps.units import LAMMPS_UNIT_CONVERSIONS


def calc_md(
    temperature=None,
    pressure=None,
    time_step=1.0,
    n_print=100,
    temperature_damping_timescale=100.0,
    pressure_damping_timescale=1000.0,
    seed=80996,
    tloop=None,
    initial_temperature=None,
    langevin=False,
    delta_temp=None,
    delta_press=None,
    rotation_matrix=None,
    units="metal",
):
    """
    Set an MD calculation within LAMMPS. Nosé Hoover is used by default.

    Args:
        temperature (None/float/list): Target temperature value(-s). If set to None, an NVE
            calculation is performed. It is required when the pressure is set or langevin is set
            It can be a list of temperature values, containing the initial target temperature and
            the final target temperature (in between the target value is varied linearly).
        pressure (None/float/numpy.ndarray/list): Target pressure. If set to None, an NVE or an
            NVT calculation is performed. If set to a scalar, the shear of the cell and the
            ratio of the x, y, and z components is kept constant, while an isotropic, hydrostatic
            pressure is applied. A list of up to length 6 can be given to specify xx, yy, zz, xy,
            xz, and yz components of the pressure tensor, respectively. These values can mix
            floats and `None` to allow only certain degrees of cell freedom to change. (Default
            is None, run isochorically.)
        n_ionic_steps (int): Number of ionic steps
        time_step (float): Step size in fs between two steps.
        n_print (int):  Print frequency
        temperature_damping_timescale (float): The time associated with the thermostat adjusting
            the temperature.  (In fs. After rescaling to appropriate time units, is equivalent to
            Lammps' `Tdamp`.)
        pressure_damping_timescale (float): The time associated with the barostat adjusting the
            temperature.  (In fs. After rescaling to appropriate time units, is equivalent to
            Lammps' `Pdamp`.)
        seed (int):  Seed for the random number generation used for the intiial velocity creation
            and langevin dynamics - otherwise ignored. If not specified, the seed is created via
            job name
        tloop:
        initial_temperature (None/float):  Initial temperature according to which the initial
            velocity field is created. If None, the initial temperature will be twice the target
            temperature (which would go immediately down to the target temperature as described
            in equipartition theorem). If 0, the velocity field is not initialized (in which case
            the initial velocity given in structure will be used). If any other number is given,
            this value is going to be used for the initial temperature.
        langevin (bool): (True or False) Activate Langevin dynamics
        delta_temp (float): Thermostat timescale, but in your Lammps time units, whatever those
            are. (DEPRECATED.)
        delta_press (float): Barostat timescale, but in your Lammps time units, whatever those
            are. (DEPRECATED.)
        job_name (str): Job name of the job to generate a unique random seed.
        rotation_matrix (numpy.ndarray): The rotation matrix from the pyiron to Lammps coordinate
            frame.
    """
    if units not in LAMMPS_UNIT_CONVERSIONS.keys():
        raise NotImplementedError
    time_units = LAMMPS_UNIT_CONVERSIONS[units]["time"]
    temperature_units = LAMMPS_UNIT_CONVERSIONS[units]["temperature"]

    # Transform time
    if time_step is not None:
        time_step *= time_units

    # Transform thermostat strength (time)
    if delta_temp is not None:
        warnings.warn(
            "WARNING: `delta_temp` is deprecated, please use `temperature_damping_timescale`."
        )
        temperature_damping_timescale = delta_temp
    else:
        temperature_damping_timescale *= time_units

    # Transform barostat strength (time)
    if delta_press is not None:
        warnings.warn(
            "WARNING: `delta_press` is deprecated, please use `pressure_damping_timescale`."
        )
        pressure_damping_timescale = delta_press
    else:
        pressure_damping_timescale *= time_units

    # Transform temperature
    if temperature is not None:
        temperature = np.array([temperature], dtype=float).flatten()
        if len(temperature) == 1:
            temperature = np.array(2 * temperature.tolist())
        elif len(temperature) != 2:
            raise ValueError(
                "At most two temperatures can be provided "
                "(for a linearly ramping target temperature), "
                "but got {}".format(len(temperature))
            )
        temperature *= temperature_units

    # Apply initial overheating (default uses the theorem of equipartition of energy between KE and PE)
    if initial_temperature is None and temperature is not None:
        initial_temperature = 2 * temperature[0]

    if seed <= 0:
        raise ValueError("Seed must be a positive integer larger than 0")

    thermo_str = ""
    # Set thermodynamic ensemble
    if pressure is not None:  # NPT
        if temperature is None or temperature.min() <= 0:
            raise ValueError(
                "Target temperature for fix nvt/npt/nph cannot be 0 or negative"
            )

        force_skewed = False
        pressure = _pressure_to_lammps(
            pressure=pressure, rotation_matrix=rotation_matrix, units=units
        )

        if np.isscalar(pressure):
            pressure_string = " iso {0} {0} {1}".format(
                pressure, pressure_damping_timescale
            )
        else:
            pressure_string = ""
            for ii, (coord, value) in enumerate(
                zip(["x", "y", "z", "xy", "xz", "yz"], pressure)
            ):
                if value is not None:
                    pressure_string += " {0} {1} {1} {2}".format(
                        coord, value, pressure_damping_timescale
                    )
                    if ii > 2:
                        force_skewed = True

        if langevin:  # NPT(Langevin)
            fix_ensemble_str = "fix ensemble all nph" + pressure_string
            thermo_str = "fix langevin all langevin {0} {1} {2} {3} zero yes".format(
                str(temperature[0]),
                str(temperature[1]),
                str(temperature_damping_timescale),
                str(seed),
            )
        else:  # NPT(Nose-Hoover)
            fix_ensemble_str = "fix ensemble all npt temp {0} {1} {2}".format(
                str(temperature[0]),
                str(temperature[1]),
                str(temperature_damping_timescale),
            )
            fix_ensemble_str += pressure_string
    elif temperature is not None:  # NVT
        if temperature.min() <= 0:
            raise ValueError(
                "Target temperature for fix nvt/npt/nph cannot be 0 or negative"
            )

        if langevin:  # NVT(Langevin)
            fix_ensemble_str = "fix ensemble all nve"
            thermo_str = "fix langevin all langevin {0} {1} {2} {3} zero yes".format(
                str(temperature[0]),
                str(temperature[1]),
                str(temperature_damping_timescale),
                str(seed),
            )
        else:  # NVT(Nose-Hoover)
            fix_ensemble_str = "fix ensemble all nvt temp {0} {1} {2}".format(
                str(temperature[0]),
                str(temperature[1]),
                str(temperature_damping_timescale),
            )
    else:  # NVE
        if langevin:
            warnings.warn("Temperature not set; Langevin ignored.")
        fix_ensemble_str = "fix ensemble all nve"

    if tloop is not None:
        fix_ensemble_str += " tloop " + str(tloop)

    line_lst = [fix_ensemble_str]
    if thermo_str != "":
        line_lst.append(thermo_str)

    line_lst.append("variable thermotime equal {} ".format(n_print))
    line_lst.append("timestep {}".format(time_step))
    if initial_temperature is not None and initial_temperature > 0:
        line_lst.append(
            _set_initial_velocity(
                temperature=initial_temperature,
                seed=seed,
                gaussian=True,
                append_value=False,
                zero_lin_momentum=True,
                zero_rot_momentum=True,
            )
        )
    line_lst += _get_thermo()
    return line_lst


def calc_minimize(
    structure: Atoms,
    ionic_energy_tolerance=0.0,
    ionic_force_tolerance=1e-4,
    max_iter=100000,
    pressure=None,
    n_print=100,
    style="cg",
    rotation_matrix=None,
    units="metal",
):
    """
    Sets parameters required for minimization.

    Args:
        ionic_energy_tolerance (float): If the magnitude of difference between energies of two consecutive steps is
            lower than or equal to `ionic_energy_tolerance`, the minimisation terminates. (Default is 0.0 eV.)
        ionic_force_tolerance (float): If the magnitude of the global force vector at a step is lower than or equal
            to `ionic_force_tolerance`, the minimisation terminates. (Default is 1e-4 eV/angstrom.)
        e_tol (float): Same as ionic_energy_tolerance (deprecated)
        f_tol (float): Same as ionic_force_tolerance (deprecated)
        max_iter (int): Maximum number of minimisation steps to carry out. If the minimisation converges before
            `max_iter` steps, terminate at the converged step. If the minimisation does not converge up to
            `max_iter` steps, terminate at the `max_iter` step. (Default is 100000.)
        pressure (None/float/numpy.ndarray/list): Target pressure. If set to None, an isochoric (constant V)
            calculation is performed. If set to a scalar, the shear of the cell and the ratio of the x, y, and
            z components is kept constant, while an isotropic, hydrostatic pressure is applied. A list of up to
            length 6 can be given to specify xx, yy, zz, xy, xz, and yz components of the pressure tensor,
            respectively. These values can mix floats and `None` to allow only certain degrees of cell freedom
            to change. (Default is None, run isochorically.)
        n_print (int): Write (dump or print) to the output file every n steps (Default: 100)
        style ('cg'/'sd'/other values from Lammps docs): The style of the numeric minimization, either conjugate
            gradient, steepest descent, or other keys permissible from the Lammps docs on 'min_style'. (Default
            is 'cg' -- conjugate gradient.)
        rotation_matrix (numpy.ndarray): The rotation matrix from the pyiron to Lammps coordinate frame.
    """
    # This docstring is a source for the calc_minimize method in pyiron_atomistics.lammps.base.LammpsBase.calc_minimize and
    # pyiron_atomistics.lammps.interactive.LammpsInteractive.calc_minimize -- Please ensure that changes to signature or
    # defaults stay consistent!

    max_evaluations = 100 * max_iter
    if n_print > max_iter:
        warnings.warn("n_print larger than max_iter, adjusting to n_print=max_iter")
        n_print = max_iter

    if units not in LAMMPS_UNIT_CONVERSIONS.keys():
        raise NotImplementedError
    energy_units = LAMMPS_UNIT_CONVERSIONS[units]["energy"]
    force_units = LAMMPS_UNIT_CONVERSIONS[units]["force"]

    ionic_energy_tolerance *= energy_units
    ionic_force_tolerance *= force_units

    line_lst = ["variable thermotime equal {} ".format(n_print)]
    line_lst += _get_thermo()
    if pressure is not None:
        if rotation_matrix is None:
            if structure is None:
                raise ValueError(
                    "No rotation matrix given while trying to convert pressure. "
                    "This is most likely due to no structure being defined."
                )
            else:
                rotation_matrix, structure = _get_rotation_matrix(
                    structure=structure, pressure=pressure
                )
        # force_skewed = False
        pressure = _pressure_to_lammps(
            pressure=pressure, rotation_matrix=rotation_matrix, units=units
        )
        if np.isscalar(pressure):
            str_press = " iso {}".format(pressure)
        else:
            str_press = ""
            for ii, (press, str_axis) in enumerate(
                zip(pressure, ["x", "y", "z", "xy", "xz", "yz"])
            ):
                if press is not None:
                    str_press += " {} {}".format(str_axis, press)
                    # if ii > 2:
                    #     force_skewed = True
            if len(str_press) > 1:
                str_press += " couple none"
        line_lst += [
            "fix ensemble all box/relax" + str_press,
        ]
    line_lst += [
        "min_style " + style,
        "minimize "
        + str(ionic_energy_tolerance)
        + " "
        + str(ionic_force_tolerance)
        + " "
        + str(int(max_iter))
        + " "
        + str(int(max_evaluations)),
    ]
    return line_lst, structure


def calc_static():
    return ["variable thermotime equal 1"] + _get_thermo() + ["run 0"]


def _get_thermo():
    return [
        "thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol",
        "thermo_modify format float %20.15g",
        "thermo ${thermotime}",
    ]


def _is_isotropic_hydrostatic(pressure):
    axial_all_alike = None not in pressure[:3] and np.allclose(
        pressure[:3], pressure[0]
    )
    shear_all_none = all(p is None for p in pressure[3:])
    shear_all_zero = None not in pressure[3:] and np.allclose(pressure[3:], 0)
    return axial_all_alike and (shear_all_none or shear_all_zero)


def _set_initial_velocity(
    temperature,
    seed=80996,
    gaussian=False,
    append_value=False,
    zero_lin_momentum=True,
    zero_rot_momentum=True,
):
    """
    Create initial velocities via velocity all create. More information can be found on LAMMPS website:
    https://lammps.sandia.gov/doc/velocity.html

    Args:
        temperature: (int or float)
        seed: (int) Seed for the initial random number generator
        gaussian: (True/False) Create velocity according to the Gaussian distribution (otherwise uniform)
        append_value: (True/False) Add the velocity values to the current velocities (probably not functional now)
        zero_lin_momentum: (True/False) Cancel the total linear momentum
        zero_rot_momentum: (True/False) Cancel the total angular momentum
    """

    arg = ""
    if gaussian:
        arg = " dist gaussian"
    if append_value:
        arg += " sum yes"
    if not zero_lin_momentum:
        arg += " mom no"
    if not zero_rot_momentum:
        arg += " rot no"
    return "velocity all create " + str(temperature) + " " + str(seed) + arg


def _pressure_to_lammps(pressure, rotation_matrix, units="metal"):
    """
    Convert a singular value, list of values, or combination of values to the appropriate six elements for Lammps
    pxx, pyy, pzz, pxy, pxz, and pyz pressure tensor representation.

    Lammps handles complex cells in a particular way, namely by using an upper triangular cell. This means we may
    need to convert our pressure tensor to a new coordinate frame. We also handle that transformation here.

    In case of a single pressure value, it is again returned as a single pressure value, to be used with the "iso"
    option (i.e., coupled deformation in x, y, and z).

    Finally, we also ensure that the units are converted from pyiron's GPa to whatever Lammps needs.

    Args:
        pressure (float/list/tuple/numpy.ndarray): The pressure(s) to convert.
        rotation_matrix (numpy.ndarray): The 3x3 matrix rotating from the pyiron to Lammps coordinate frame.

    Returns:
        (list): pxx, pyy, pzz, pxy, pxz, and pyz to be passed to Lammps.

        or

        (float): a single, isotropic pressure to be used with the "iso" option
    """

    # If pressure is a scalar, only unit conversion is needed.
    if np.isscalar(pressure):
        return float(pressure) * LAMMPS_UNIT_CONVERSIONS[units]["pressure"]

    # Normalize pressure to a list of 6 entries of either float or NoneType type.
    if len(pressure) > 6:
        raise ValueError(
            "Pressure can have a maximum of 6 values, (x, y, z, xy, xz, and yz), but got "
            + "{}".format(len(pressure))
        )
    pressure = [float(p) if p is not None else None for p in pressure]
    pressure += (6 - len(pressure)) * [None]

    if all(p is None for p in pressure):
        raise ValueError("Pressure cannot have a length but all be None")

    # If necessary, rotate the pressure tensor to the Lammps coordinate frame.
    # Isotropic, hydrostatic pressures are rotation invariant.
    if not np.isclose(
        np.matrix.trace(rotation_matrix), 3
    ) and not _is_isotropic_hydrostatic(pressure):
        if any(p is None for p in pressure):
            raise ValueError(
                "Cells which are not orthorhombic or an upper-triangular cell are incompatible with Lammps "
                "constant pressure calculations unless the entire pressure tensor is defined. "
                "The reason is that Lammps demands such cells be represented with an "
                "upper-triangular unit cell, thus a rotation between Lammps and pyiron coordinate "
                "frames is required; it is not possible to rotate the pressure tensor if any of "
                "its components is None."
            )
        pxx, pyy, pzz, pxy, pxz, pyz = pressure
        pressure_tensor = np.array([[pxx, pxy, pxz], [pxy, pyy, pyz], [pxz, pyz, pzz]])
        lammps_pressure_tensor = rotation_matrix.T @ pressure_tensor @ rotation_matrix
        pressure = list(lammps_pressure_tensor[[0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]])

    return [
        (p * LAMMPS_UNIT_CONVERSIONS[units]["pressure"] if p is not None else p)
        for p in pressure
    ]


def _get_rotation_matrix(structure, pressure):
    """

    Args:
        pressure:

    Returns:

    """
    if structure is not None:
        prism = UnfoldingPrism(structure.cell)

        structure = _modify_structure_to_allow_requested_deformation(
            pressure=pressure, structure=structure, prism=prism
        )
        rotation_matrix = prism.R
    else:
        warnings.warn("No structure set, can not validate the simulation cell!")
        rotation_matrix = None
    return rotation_matrix, structure


def _modify_structure_to_allow_requested_deformation(structure, pressure, prism=None):
    """
    Lammps will not allow xy/xz/yz cell deformations in minimization or MD for non-triclinic cells. In case the
    requested pressure for a calculation has these non-diagonal entries, we need to make sure it will run. One way
    to do this is by invoking the lammps `change_box` command, but it is easier to just force our box to to be
    triclinic by adding a very small cell perturbation (in the case where it isn't triclinic already).

    Args:
        pressure (float/int/list/numpy.ndarray/tuple): Between three and six pressures for the x, y, z, xy, xz, and
            yz directions, in that order, or a single value.
    """
    if hasattr(pressure, "__len__"):
        non_diagonal_pressures = np.any([p is not None for p in pressure[3:]])

        if prism is None:
            prism = UnfoldingPrism(structure.cell)

        if non_diagonal_pressures:
            try:
                if not prism.is_skewed():
                    skew_structure = structure.copy()
                    skew_structure.cell[0, 1] += 2 * prism.acc
                    return skew_structure
            except AttributeError:
                warnings.warn(
                    "WARNING: Setting a calculation type which uses pressure before setting the structure risks "
                    + "constraining your cell shape evolution if non-diagonal pressures are used but the structure "
                    + "is not triclinic from the start of the calculation."
                )
    return structure



import os
import shutil
import subprocess
from typing import Optional

import pandas
from ase.atoms import Atoms

from pyiron_lammps.compatibility.calculate import (
    calc_md,
    calc_minimize,
    calc_static,
)
from pyiron_lammps.compatibility.constraints import set_selective_dynamics
from pyiron_lammps.output import parse_lammps_output
from pyiron_lammps.potential import get_potential_by_name
from pyiron_lammps.structure import write_lammps_datafile


def lammps_file_interface_function(
    working_directory: str,
    structure: Atoms,
    potential: str,
    calc_mode: str = "static",
    calc_kwargs: Optional[dict] = None,
    units: str = "metal",
    lmp_command: str = "mpiexec -n 1 --oversubscribe lmp_mpi -in lmp.in",
    resource_path: Optional[str] = None,
    input_control_file: Optional[dict] = None,
    write_restart_file: bool = False,
    read_restart_file: bool = False,
    restart_file: str = "restart.out",
):
    """
    A single function to execute a LAMMPS calculation based on the LAMMPS job implemented in pyiron

    Examples:

    >>> import os
    >>> from ase.build import bulk
    >>> from pyiron_atomistics.lammps.lammps import lammps_function
    >>>
    >>> shell_output, parsed_output, job_crashed = lammps_function(
    ...     working_directory=os.path.abspath("lmp_working_directory"),
    ...     structure=bulk("Al", cubic=True),
    ...     potential='2009--Mendelev-M-I--Al-Mg--LAMMPS--ipr1',
    ...     calc_mode="md",
    ...     calc_kwargs={"temperature": 500.0, "pressure": 0.0, "n_ionic_steps": 1000, "n_print": 100},
    ...     cutoff_radius=None,
    ...     units="metal",
    ...     bonds_kwargs={},
    ...      enable_h5md=False,
    ... )

    Args:
        working_directory (str): directory in which the LAMMPS calculation is executed
        structure (Atoms): ase.atoms.Atoms - atomistic structure
        potential (str): Name of the LAMMPS potential based on the NIST database and the OpenKIM database
        calc_mode (str): select mode of calculation ["static", "md", "minimize", "vcsgc"]
        calc_kwargs (dict): key-word arguments for the calculate function, the input parameters depend on the calc_mode:
          "static": No parameters
          "md": "temperature", "pressure", "n_ionic_steps", "time_step", "n_print", "temperature_damping_timescale",
                "pressure_damping_timescale", "seed", "tloop", "initial_temperature", "langevin", "delta_temp",
                "delta_press", job_name", "rotation_matrix"
          "minimize": "ionic_energy_tolerance", "ionic_force_tolerance", "max_iter", "pressure", "n_print", "style",
                      "rotation_matrix"
          "vcsgc": "mu", "ordered_element_list", "target_concentration", "kappa", "mc_step_interval", "swap_fraction",
                   "temperature_mc", "window_size", "window_moves", "temperature", "pressure", "n_ionic_steps",
                   "time_step", "n_print", "temperature_damping_timescale", "pressure_damping_timescale", "seed",
                   "initial_temperature", "langevin", "job_name", "rotation_matrix"
        cutoff_radius (float): cut-off radius for the interatomic potential
        units (str): Units for LAMMPS
        bonds_kwargs (dict): key-word arguments to create atomistic bonds:
          "species", "element_list", "cutoff_list", "max_bond_list", "bond_type_list", "angle_type_list",
        server_kwargs (dict): key-word arguments to create server object - the available parameters are:
          "user", "host", "run_mode", "queue", "qid", "cores", "threads", "new_h5", "structure_id", "run_time",
          "memory_limit", "accept_crash", "additional_arguments", "gpus", "conda_environment_name",
          "conda_environment_path"
        enable_h5md (bool): activate h5md mode for LAMMPS
        write_restart_file (bool): enable writing the LAMMPS restart file
        read_restart_file (bool): enable loading the LAMMPS restart file
        restart_file (str): file name of the LAMMPS restart file to copy
        executable_version (str): LAMMPS version to for the execution
        executable_path (str): path to the LAMMPS executable
        input_control_file (str|list|dict): Option to modify the LAMMPS input file directly

    Returns:
        str, dict, bool: Tuple consisting of the shell output (str), the parsed output (dict) and a boolean flag if
                         the execution raised an accepted error.
    """
    if calc_kwargs is None:
        calc_kwargs = {}
    else:
        calc_kwargs = calc_kwargs.copy()

    os.makedirs(working_directory, exist_ok=True)
    potential_lst, potential_replace, species = _get_potential(
        potential=potential, resource_path=resource_path
    )

    lmp_str_lst = []
    atom_type = "atomic"
    for l in lammps_file_initialization(
        structure=structure,
        units=units,
        read_restart_file=read_restart_file,
        restart_file=restart_file,
    ):
        if l.startswith("units") and "units" in potential_replace:
            lmp_str_lst.append(potential_replace["units"])
        elif l.startswith("atom_style") and "atom_style" in potential_replace:
            lmp_str_lst.append(potential_replace["atom_style"])
            atom_type = potential_replace["atom_style"].split()[-1]
        elif l.startswith("dimension") and "dimension" in potential_replace:
            lmp_str_lst.append(potential_replace["dimension"])
        else:
            lmp_str_lst.append(l)

    lmp_str_lst += potential_lst
    lmp_str_lst += ["variable dumptime equal {} ".format(calc_kwargs.get("n_print", 1))]
    lmp_str_lst += [
        "dump 1 all custom ${dumptime} dump.out id type xsu ysu zsu fx fy fz vx vy vz",
        'dump_modify 1 sort id format line "%d %d %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g"',
    ]

    if calc_mode == "static":
        lmp_str_lst += [
            k + " " + v
            for k, v in set_selective_dynamics(
                structure=structure, calc_md=False
            ).items()
        ]
        lmp_str_tmp_lst = calc_static()
        lmp_str_lst = _modify_input_dict(
            input_control_file=input_control_file,
            lmp_str_lst=lmp_str_lst + lmp_str_tmp_lst[:-1],
        )
        lmp_str_lst.append(lmp_str_tmp_lst[-1])
    elif calc_mode == "md":
        lmp_str_lst += [
            k + " " + v
            for k, v in set_selective_dynamics(
                structure=structure, calc_md=True
            ).items()
        ]
        if "n_ionic_steps" in calc_kwargs.keys():
            n_ionic_steps = int(calc_kwargs.pop("n_ionic_steps"))
        else:
            n_ionic_steps = 1
        if read_restart_file:
            calc_kwargs["initial_temperature"] = 0.0
        calc_kwargs["units"] = units
        lmp_str_lst += calc_md(**calc_kwargs)
        lmp_str_lst = _modify_input_dict(
            input_control_file=input_control_file,
            lmp_str_lst=lmp_str_lst,
        )
        if read_restart_file:
            lmp_str_lst += ["reset_timestep 0"]
        lmp_str_lst += ["run {} ".format(n_ionic_steps)]
    elif calc_mode == "minimize":
        calc_kwargs["units"] = units
        lmp_str_lst += [
            k + " " + v
            for k, v in set_selective_dynamics(
                structure=structure, calc_md=False
            ).items()
        ]
        lmp_str_tmp_lst, structure = calc_minimize(structure=structure, **calc_kwargs)
        lmp_str_lst = _modify_input_dict(
            input_control_file=input_control_file,
            lmp_str_lst=lmp_str_lst + lmp_str_tmp_lst[:-1],
        )
        lmp_str_lst.append(lmp_str_tmp_lst[-1])
    else:
        raise ValueError(
            f"calc_mode must be one of: static, md or minimize, not {calc_mode}"
        )
    if read_restart_file:
        shutil.copyfile(
            os.path.abspath(restart_file),
            os.path.join(working_directory, os.path.basename(restart_file)),
        )
    if write_restart_file:
        lmp_str_lst.append(f"write_restart {os.path.basename(restart_file)}")

    with open(os.path.join(working_directory, "lmp.in"), "w") as f:
        f.writelines([l + "\n" for l in lmp_str_lst])

    write_lammps_datafile(
        structure=structure,
        potential_elements=species,
        bond_dict=None,
        units=units,
        file_name="lammps.data",
        working_directory=working_directory,
        atom_type=atom_type,
    )

    shell = subprocess.check_output(
        lmp_command,
        cwd=working_directory,
        shell=True,
        universal_newlines=True,
        env=os.environ.copy(),
    )
    output = parse_lammps_output(
        working_directory=working_directory,
        structure=structure,
        potential_elements=species,
        units=units,
        prism=None,
        dump_h5_file_name="dump.h5",
        dump_out_file_name="dump.out",
        log_lammps_file_name="log.lammps",
    )
    return shell, output, False


def lammps_file_initialization(
    structure, dimension=3, units="metal", read_restart_file=False, restart_file=None
):
    init_commands = ["units " + units]
    boundary = " ".join(["p" if coord else "f" for coord in structure.pbc])
    if read_restart_file:
        init_commands.append(f"read_restart {os.path.basename(restart_file)}")
    else:
        init_commands += [
            "dimension " + str(dimension),
            "boundary " + boundary + "",
            "atom_style atomic",
            "read_data lammps.data",
        ]
    return init_commands


def _modify_input_dict(
    input_control_file: Optional[dict] = None,
    lmp_str_lst: list[str] = [],
):
    if input_control_file is not None:
        lmp_tmp_lst, keys_used = [], []
        for l in lmp_str_lst:
            ls = l.split()
            if len(ls) >= 1:  # Remove empty lines
                key = ls[0]
                if key in input_control_file.keys():
                    lmp_tmp_lst.append(key + " " + input_control_file[key])
                    keys_used.append(key)
                else:
                    lmp_tmp_lst.append(l)
        for k, v in input_control_file.items():
            if k not in keys_used:
                lmp_tmp_lst.append(k + " " + v)

        return lmp_tmp_lst
    else:
        return lmp_str_lst


def _get_potential(potential, resource_path: Optional[str] = None):
    if isinstance(potential, str):
        potential_dataframe = get_potential_by_name(
            potential_name=potential, resource_path=resource_path
        )
    elif isinstance(potential, pandas.DataFrame):
        potential_dataframe = potential.iloc[0]
    elif isinstance(potential, pandas.Series):
        potential_dataframe = potential
    else:
        raise TypeError()

    potential_replace = {}
    potential_lst = []
    for l in potential_dataframe["Config"]:
        if l.startswith("units"):
            potential_replace["units"] = l
        elif l.startswith("atom_style"):
            potential_replace["atom_style"] = l
        elif l.startswith("dimension"):
            potential_replace["dimension"] = l
        else:
            potential_lst.append(l)

    return potential_lst, potential_replace, potential_dataframe["Species"]



from __future__ import annotations

import os
import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ase.atoms import Atoms

from pyiron_lammps.output_raw import (
    parse_raw_dump_from_h5md,
    parse_raw_dump_from_text,
    parse_raw_lammps_log,
)
from pyiron_lammps.structure import UnfoldingPrism
from pyiron_lammps.units import UnitConverter


def remap_indices_ase(
    lammps_indices: Union[np.ndarray, List],
    potential_elements: Union[np.ndarray, List],
    structure: Atoms,
) -> np.ndarray:
    """
    Give the Lammps-dumped indices, re-maps these back onto the structure's indices to preserve the species.

    The issue is that for an N-element potential, Lammps dumps the chemical index from 1 to N based on the order
    that these species are written in the Lammps input file. But the indices for a given structure are based on the
    order in which chemical species were added to that structure, and run from 0 up to the number of species
    currently in that structure. Therefore we need to be a little careful with mapping.

    Args:
        lammps_indices (numpy.ndarray/list): The Lammps-dumped integers.
        potential_elements (numpy.ndarray/list):
        structure (pyiron_atomistics.atomistics.structure.Atoms):

    Returns:
        numpy.ndarray: Those integers mapped onto the structure.
    """
    lammps_symbol_order = np.array(potential_elements)

    # Create a map between the lammps indices and structure indices to preserve species
    structure_symbol_order = np.unique(structure.get_chemical_symbols())
    map_ = np.array(
        [
            int(np.argwhere(lammps_symbol_order == symbol)[0][0]) + 1
            for symbol in structure_symbol_order
        ]
    )

    structure_indices = np.array(lammps_indices)
    for i_struct, i_lammps in enumerate(map_):
        np.place(structure_indices, lammps_indices == i_lammps, i_struct)
    # TODO: Vectorize this for-loop for computational efficiency

    return structure_indices


def parse_lammps_output(
    working_directory: str,
    structure: Atoms,
    potential_elements: Union[np.ndarray, List],
    units: str,
    prism: Optional[UnfoldingPrism] = None,
    dump_h5_file_name: str = "dump.h5",
    dump_out_file_name: str = "dump.out",
    log_lammps_file_name: str = "log.lammps",
    remap_indices_funct: callable = remap_indices_ase,
) -> Dict:
    if prism is None:
        prism = UnfoldingPrism(structure.cell)
    dump_dict = _parse_dump(
        dump_h5_full_file_name=os.path.join(working_directory, dump_h5_file_name),
        dump_out_full_file_name=os.path.join(working_directory, dump_out_file_name),
        prism=prism,
        structure=structure,
        potential_elements=potential_elements,
        remap_indices_funct=remap_indices_funct,
    )

    generic_keys_lst, pressure_dict, df = _parse_log(
        log_lammps_full_file_name=os.path.join(working_directory, log_lammps_file_name),
        prism=prism,
    )

    convert_units = UnitConverter(units).convert_array_to_pyiron_units

    hdf_output = {"generic": {}, "lammps": {}}
    hdf_generic = hdf_output["generic"]
    hdf_lammps = hdf_output["lammps"]

    if "computes" in dump_dict.keys():
        for k, v in dump_dict.pop("computes").items():
            hdf_generic[k] = convert_units(np.array(v), label=k)

    hdf_generic["steps"] = convert_units(
        np.array(dump_dict.pop("steps"), dtype=int), label="steps"
    )

    for k, v in dump_dict.items():
        if len(v) > 0:
            try:
                hdf_generic[k] = convert_units(np.array(v), label=k)
            except ValueError:
                hdf_generic[k] = [convert_units(np.array(val), label=k) for val in v]

    if df is not None and pressure_dict is not None and generic_keys_lst is not None:
        for k, v in df.items():
            v = convert_units(np.array(v), label=k)
            if k in generic_keys_lst:
                hdf_generic[k] = v
            else:  # This is a hack for backward comparability
                hdf_lammps[k] = v

        # Store pressures as numpy arrays
        for key, val in pressure_dict.items():
            hdf_generic[key] = convert_units(val, label=key)
    else:
        warnings.warn("LAMMPS warning: No log.lammps output file found.")

    return hdf_output


def _parse_dump(
    dump_h5_full_file_name: str,
    dump_out_full_file_name: str,
    prism: UnfoldingPrism,
    structure: Atoms,
    potential_elements: Union[np.ndarray, List],
    remap_indices_funct: callable = remap_indices_ase,
) -> Dict:
    if os.path.isfile(dump_h5_full_file_name):
        if not _check_ortho_prism(prism=prism):
            raise RuntimeError(
                "The Lammps output will not be mapped back to pyiron correctly."
            )
        return parse_raw_dump_from_h5md(
            file_name=dump_h5_full_file_name,
        )
    elif os.path.exists(dump_out_full_file_name):
        return _collect_dump_from_text(
            file_name=dump_out_full_file_name,
            prism=prism,
            structure=structure,
            potential_elements=potential_elements,
            remap_indices_funct=remap_indices_funct,
        )
    else:
        raise FileNotFoundError(
            f"Neither {dump_h5_full_file_name} nor {dump_out_full_file_name} exist."
        )


def _collect_dump_from_text(
    file_name: str,
    prism: UnfoldingPrism,
    structure: Atoms,
    potential_elements: Union[np.ndarray, List],
    remap_indices_funct: callable = remap_indices_ase,
) -> Dict:
    """
    general purpose routine to extract static from a lammps dump file
    """
    rotation_lammps2orig = prism.R.T
    dump_lammps_dict = parse_raw_dump_from_text(file_name=file_name)
    dump_dict = {}
    for key, val in dump_lammps_dict.items():
        if key in ["cells"]:
            dump_dict[key] = [prism.unfold_cell(cell=cell) for cell in val]
        elif key in ["indices"]:
            dump_dict[key] = [
                remap_indices_funct(
                    lammps_indices=indices,
                    potential_elements=potential_elements,
                    structure=structure,
                )
                for indices in val
            ]
        elif key in [
            "forces",
            "mean_forces",
            "velocities",
            "mean_velocities",
            "mean_unwrapped_positions",
        ]:
            dump_dict[key] = [np.matmul(v, rotation_lammps2orig) for v in val]
        elif key in ["positions", "unwrapped_positions"]:
            dump_dict[key] = [
                np.matmul(np.matmul(v, lammps_cell), rotation_lammps2orig)
                for v, lammps_cell in zip(val, dump_lammps_dict["cells"])
            ]
        else:
            dump_dict[key] = val
    return dump_dict


def _parse_log(
    log_lammps_full_file_name: str, prism: UnfoldingPrism
) -> Union[Tuple[List[str], Dict, pd.DataFrame], Tuple[None, None, None]]:
    """
    If it exists, parses the lammps log file and either raises an exception if errors
    occurred or returns data. Just returns a tuple of Nones if there is no file at the
    given location.

    Args:
        log_lammps_full_file_name (str): The path to the lammps log file.
        prism (pyiron_atomistics.lammps.structure.UnfoldingPrism): For mapping between
            lammps and pyiron structures

    Returns:
        (list | None): Generic keys
        (dict | None): Pressures
        (pandas.DataFrame | None): A dataframe with the rest of the information

    Raises:
        (RuntimeError): If there are "ERROR" tags in the log.
    """
    if os.path.exists(log_lammps_full_file_name):
        return _collect_output_log(
            file_name=log_lammps_full_file_name,
            prism=prism,
        )
    else:
        return None, None, None


def _collect_output_log(
    file_name: str, prism: UnfoldingPrism
) -> Tuple[List[str], Dict, pd.DataFrame]:
    """
    general purpose routine to extract static from a lammps log file
    """
    df = parse_raw_lammps_log(file_name=file_name)

    h5_dict = {
        "Step": "steps",
        "Temp": "temperature",
        "PotEng": "energy_pot",
        "TotEng": "energy_tot",
        "Volume": "volume",
        "LogStep": "LogStep",
    }
    if "LogStep" not in df.columns:
        del h5_dict["LogStep"]

    for key in df.columns[df.columns.str.startswith("f_mean")]:
        h5_dict[key] = key.replace("f_", "")

    df = df.rename(index=str, columns=h5_dict)
    pressure_dict = dict()
    if all(
        [
            x in df.columns.values
            for x in [
                "Pxx",
                "Pxy",
                "Pxz",
                "Pxy",
                "Pyy",
                "Pyz",
                "Pxz",
                "Pyz",
                "Pzz",
            ]
        ]
    ):
        pressures = (
            np.stack(
                (
                    df.Pxx,
                    df.Pxy,
                    df.Pxz,
                    df.Pxy,
                    df.Pyy,
                    df.Pyz,
                    df.Pxz,
                    df.Pyz,
                    df.Pzz,
                ),
                axis=-1,
            )
            .reshape(-1, 3, 3)
            .astype("float64")
        )
        # Rotate pressures from Lammps frame to pyiron frame if necessary
        if _check_ortho_prism(prism=prism):
            rotation_matrix = prism.R.T
            pressures = rotation_matrix.T @ pressures @ rotation_matrix

        df = df.drop(
            columns=df.columns[
                ((df.columns.str.len() == 3) & df.columns.str.startswith("P"))
            ]
        )
        pressure_dict["pressures"] = pressures
    else:
        warnings.warn(
            "LAMMPS warning: log.lammps does not contain the required pressure values."
        )
    if "mean_pressure[1]" in df.columns:
        pressures = (
            np.stack(
                tuple(df[f"mean_pressure[{i}]"] for i in [1, 4, 5, 4, 2, 6, 5, 6, 3]),
                axis=-1,
            )
            .reshape(-1, 3, 3)
            .astype("float64")
        )
        if _check_ortho_prism(prism=prism):
            rotation_matrix = prism.R.T
            pressures = rotation_matrix.T @ pressures @ rotation_matrix
        df = df.drop(
            columns=df.columns[
                (
                    df.columns.str.startswith("mean_pressure")
                    & df.columns.str.endswith("]")
                )
            ]
        )
        pressure_dict["mean_pressures"] = pressures
    generic_keys_lst = list(h5_dict.values())
    return generic_keys_lst, pressure_dict, df


def _check_ortho_prism(
    prism: UnfoldingPrism, rtol: float = 0.0, atol: float = 1e-08
) -> bool:
    """
    Check if the rotation matrix of the UnfoldingPrism object is sufficiently close to a unit matrix

    Args:
        prism (pyiron_atomistics.lammps.structure.UnfoldingPrism): UnfoldingPrism object to check
        rtol (float): relative precision for numpy.isclose()
        atol (float): absolute precision for numpy.isclose()

    Returns:
        boolean: True or False
    """
    return np.isclose(prism.R, np.eye(3), rtol=rtol, atol=atol).all()



from __future__ import annotations

import warnings
from dataclasses import asdict, dataclass, field
from io import StringIO
from typing import Dict, List, Union

import numpy as np
import pandas as pd


@dataclass
class DumpData:
    steps: List = field(default_factory=lambda: [])
    natoms: List = field(default_factory=lambda: [])
    cells: List = field(default_factory=lambda: [])
    indices: List = field(default_factory=lambda: [])
    forces: List = field(default_factory=lambda: [])
    mean_forces: List = field(default_factory=lambda: [])
    velocities: List = field(default_factory=lambda: [])
    mean_velocities: List = field(default_factory=lambda: [])
    unwrapped_positions: List = field(default_factory=lambda: [])
    mean_unwrapped_positions: List = field(default_factory=lambda: [])
    positions: List = field(default_factory=lambda: [])
    computes: Dict = field(default_factory=lambda: {})


def to_amat(l_list: Union[np.ndarray, List]) -> List:
    lst = np.reshape(l_list, -1)
    if len(lst) == 9:
        (
            xlo_bound,
            xhi_bound,
            xy,
            ylo_bound,
            yhi_bound,
            xz,
            zlo_bound,
            zhi_bound,
            yz,
        ) = lst

    elif len(lst) == 6:
        xlo_bound, xhi_bound, ylo_bound, yhi_bound, zlo_bound, zhi_bound = lst
        xy, xz, yz = 0.0, 0.0, 0.0
    else:
        raise ValueError("This format for amat not yet implemented: " + str(len(lst)))

    # > xhi_bound - xlo_bound = xhi -xlo  + MAX(0.0, xy, xz, xy + xz) - MIN(0.0, xy, xz, xy + xz)
    # > xhili = xhi -xlo   = xhi_bound - xlo_bound - MAX(0.0, xy, xz, xy + xz) + MIN(0.0, xy, xz, xy + xz)
    xhilo = (
        (xhi_bound - xlo_bound)
        - max([0.0, xy, xz, xy + xz])
        + min([0.0, xy, xz, xy + xz])
    )

    # > yhilo = yhi -ylo = yhi_bound -ylo_bound - MAX(0.0, yz) + MIN(0.0, yz)
    yhilo = (yhi_bound - ylo_bound) - max([0.0, yz]) + min([0.0, yz])

    # > zhi - zlo = zhi_bound- zlo_bound
    zhilo = zhi_bound - zlo_bound

    cell = [[xhilo, 0, 0], [xy, yhilo, 0], [xz, yz, zhilo]]
    return cell


def parse_raw_dump_from_h5md(file_name: str) -> Dict:
    import h5py

    with h5py.File(file_name, mode="r", libver="latest", swmr=True) as h5md:
        positions = [pos_i.tolist() for pos_i in h5md["/particles/all/position/value"]]
        steps = [steps_i.tolist() for steps_i in h5md["/particles/all/position/step"]]
        forces = [for_i.tolist() for for_i in h5md["/particles/all/force/value"]]
        # following the explanation at: http://nongnu.org/h5md/h5md.html
        cell = [
            np.eye(3) * np.array(cell_i.tolist())
            for cell_i in h5md["/particles/all/box/edges/value"]
        ]
    return {
        "forces": forces,
        "positions": positions,
        "steps": steps,
        "cells": cell,
    }


def parse_raw_dump_from_text(file_name: str) -> Dict:
    """
    Docstring for _parse_dump_from_text

    Args:
        file_name (str): The path to the lammps dump file.

    Returns:
        Dict: Parsed dump data.
    """
    with open(file_name, "r") as f:
        dump = DumpData()

        for line in f:
            if "ITEM: TIMESTEP" in line:
                dump.steps.append(int(f.readline()))

            elif "ITEM: BOX BOUNDS" in line:
                c1 = np.fromstring(f.readline(), dtype=float, sep=" ")
                c2 = np.fromstring(f.readline(), dtype=float, sep=" ")
                c3 = np.fromstring(f.readline(), dtype=float, sep=" ")
                cell = np.concatenate([c1, c2, c3])
                dump.cells.append(to_amat(cell))

            elif "ITEM: NUMBER OF ATOMS" in line:
                n = int(f.readline())
                dump.natoms.append(n)

            elif "ITEM: ATOMS" in line:
                # get column names from line
                columns = line.lstrip("ITEM: ATOMS").split()

                # Read line by line of snapshot into a string buffer
                # Than parse using pandas for speed and column acces
                buf = StringIO()
                for _ in range(n):
                    buf.write(f.readline())
                buf.seek(0)
                df = pd.read_csv(
                    buf,
                    nrows=n,
                    sep="\\s+",
                    header=None,
                    names=columns,
                    engine="c",
                )
                df.sort_values(by="id", ignore_index=True, inplace=True)
                # Coordinate transform lammps->pyiron
                dump.indices.append(df["type"].array.astype(int))

                dump.forces.append(
                    np.stack([df["fx"].array, df["fy"].array, df["fz"].array], axis=1)
                )
                if "f_mean_forces[1]" in columns:
                    dump.mean_forces.append(
                        np.stack(
                            [
                                df["f_mean_forces[1]"].array,
                                df["f_mean_forces[2]"].array,
                                df["f_mean_forces[3]"].array,
                            ],
                            axis=1,
                        )
                    )
                if "vx" in columns and "vy" in columns and "vz" in columns:
                    dump.velocities.append(
                        np.stack(
                            [
                                df["vx"].array,
                                df["vy"].array,
                                df["vz"].array,
                            ],
                            axis=1,
                        )
                    )

                if "f_mean_velocities[1]" in columns:
                    dump.mean_velocities.append(
                        np.stack(
                            [
                                df["f_mean_velocities[1]"].array,
                                df["f_mean_velocities[2]"].array,
                                df["f_mean_velocities[3]"].array,
                            ],
                            axis=1,
                        )
                    )

                if "xsu" in columns:
                    direct_unwrapped_positions = np.stack(
                        [
                            df["xsu"].array,
                            df["ysu"].array,
                            df["zsu"].array,
                        ],
                        axis=1,
                    )
                    dump.unwrapped_positions.append(direct_unwrapped_positions)
                    dump.positions.append(
                        direct_unwrapped_positions
                        - np.floor(direct_unwrapped_positions)
                    )

                if "f_mean_positions[1]" in columns:
                    dump.mean_unwrapped_positions.append(
                        np.stack(
                            [
                                df["f_mean_positions[1]"].array,
                                df["f_mean_positions[2]"].array,
                                df["f_mean_positions[3]"].array,
                            ],
                            axis=1,
                        )
                    )
                for k in columns:
                    if k.startswith("c_"):
                        kk = k.replace("c_", "")
                        if kk not in dump.computes.keys():
                            dump.computes[kk] = []
                        dump.computes[kk].append(df[k].array)

        return asdict(dump)


def parse_raw_lammps_log(file_name: str) -> pd.DataFrame:
    """
    Docstring for _parse_lammps_log

    Args:
        file_name (str): The path to the lammps log file.

    Returns:
        pd.DataFrame: Dataframe containing the parsed log data.
    """
    with open(file_name, "r") as f:
        dfs = []
        read_thermo = False
        for l in f:
            l = l.lstrip()

            if l.startswith("Step"):
                thermo_lines = ""
                read_thermo = True

            if read_thermo:
                if l.startswith("Loop") or l.startswith("ERROR"):
                    read_thermo = False
                    dfs.append(
                        pd.read_csv(StringIO(thermo_lines), sep="\\s+", engine="c")
                    )

                elif l.startswith("WARNING:"):
                    warnings.warn(f"A warning was found in the log:\n{l}")

                else:
                    thermo_lines += l

    if len(dfs) == 1:
        df = dfs[0]
    else:
        for i in range(len(dfs)):
            df = dfs[i]
            df["LogStep"] = np.ones(len(df)) * i
        df = pd.concat(dfs, ignore_index=True)
    return df



import os
from pathlib import Path
from typing import Optional, Union

import pandas
from ase.atoms import Atoms

potential_installation = """
Potential installation guide:

1. Check whether iprpy-data is installed. If not, install it using:

`conda install -c conda-forge iprpy-data`

2. Check whether the resource path is set via:

```python
import os
print(os.environ["CONDA_PREFIX"])
```

3. If the resource path is set, you can call the potential using:

```python
from atomistics.calculators import get_potential_by_name


get_potential_by_name(
    potential_name=my_potential,
    resource_path=os.path.join(os.environ["CONDA_PREFIX"], "share", "iprpy"),
)
```

"""


class PotentialAbstract:
    """
    The PotentialAbstract class loads a list of available potentials and sorts them. Afterwards the potentials can be
    accessed through:
        PotentialAbstract.<Element>.<Element> or PotentialAbstract.find_potentials_set({<Element>, <Element>}

    Args:
        potential_df:
        default_df:
        selected_atoms:
    """

    def __init__(
        self,
        potential_df: pandas.DataFrame,
        default_df: pandas.DataFrame = None,
        selected_atoms: list[str] = None,
    ):
        self._potential_df = potential_df
        self._default_df = default_df
        if selected_atoms is not None:
            self._selected_atoms = selected_atoms
        else:
            self._selected_atoms = []

    def find(self, element: Union[set[str], list[str], str]) -> pandas.DataFrame:
        """
        Find the potentials

        Args:
            element (set, str): element or set of elements for which you want the possible LAMMPS potentials

        Returns:
            list: of possible potentials for the element or the combination of elements

        """
        if isinstance(element, list):
            element = set(element)
        elif isinstance(element, str):
            element = {element}
        elif not isinstance(element, set):
            raise TypeError("Only, str, list and set supported!")
        return self._potential_df[
            [
                bool(set(element).issubset(species))
                for species in self._potential_df["Species"].values
            ]
        ]

    def find_by_name(self, potential_name: str) -> pandas.DataFrame:
        mask = self._potential_df["Name"] == potential_name
        if not mask.any():
            raise ValueError(f"Potential '{potential_name}' not found in database.")
        return self._potential_df[mask]

    def list(self) -> pandas.DataFrame:
        """
        List the available potentials

        Returns:
            list: of possible potentials for the element or the combination of elements
        """
        return self._potential_df

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        potential_df = self.find(element=item)
        selected_atoms = self._selected_atoms + [item]
        return PotentialAbstract(
            potential_df=potential_df,
            default_df=self._default_df,
            selected_atoms=selected_atoms,
        )

    def __str__(self) -> str:
        return str(self.list())

    @staticmethod
    def _get_potential_df(file_name_lst, resource_path):
        """

        Args:
            file_name_lst (set):
            resource_path (str):

        Returns:
            pandas.DataFrame:
        """
        for path, _folder_lst, file_lst in os.walk(resource_path):
            for periodic_table_file_name in file_name_lst:
                if (
                    periodic_table_file_name in file_lst
                    and periodic_table_file_name.endswith(".csv")
                ):
                    return pandas.read_csv(
                        os.path.join(path, periodic_table_file_name),
                        index_col=0,
                        converters={
                            "Species": (
                                lambda x: x.replace("'", "").strip("[]").split(", ")
                            ),
                            "Config": (
                                lambda x: x.replace("'", "")
                                .replace("\\n", "\n")
                                .strip("[]")
                                .split(", ")
                            ),
                            "Filename": (
                                lambda x: x.replace("'", "").strip("[]").split(", ")
                            ),
                        },
                    )
        raise ValueError(
            "Was not able to locate the potential files." + potential_installation
        )


class LammpsPotentialFile(PotentialAbstract):
    """
    The Potential class is derived from the PotentialAbstract class, but instead of loading the potentials from a list,
    the potentials are loaded from a file.

    Args:
        potential_df:
        default_df:
        selected_atoms:
    """

    def __init__(
        self,
        potential_df=None,
        default_df=None,
        selected_atoms=None,
        resource_path=None,
    ):
        if potential_df is None:
            potential_df = self._get_potential_df(
                file_name_lst={"potentials_lammps.csv"},
                resource_path=resource_path,
            )
        super().__init__(
            potential_df=potential_df,
            default_df=default_df,
            selected_atoms=selected_atoms,
        )
        self._resource_path = resource_path

    def default(self):
        if self._default_df is not None:
            atoms_str = "_".join(sorted(self._selected_atoms))
            return self._default_df[
                (self._default_df["Name"] == self._default_df.loc[atoms_str].values[0])
            ]
        return None

    def find_default(
        self, element: Union[set[str], list[str], str]
    ) -> pandas.DataFrame:
        """
        Find the potentials

        Args:
            element (set, str): element or set of elements for which you want the possible LAMMPS potentials
            path (bool): choose whether to return the full path to the potential or just the potential name

        Returns:
            list: of possible potentials for the element or the combination of elements

        """
        if isinstance(element, list):
            element = set(element)
        elif isinstance(element, str):
            element = {element}
        elif not isinstance(element, set):
            raise TypeError("Only, str, list and set supported!")
        element_lst = list(element)
        if self._default_df is not None:
            merged_lst = list(set(self._selected_atoms + element_lst))
            atoms_str = "_".join(sorted(merged_lst))
            return self._default_df[
                (self._default_df["Name"] == self._default_df.loc[atoms_str].values[0])
            ]
        return None

    def __getitem__(self, item):
        potential_df = self.find(element=item)
        selected_atoms = self._selected_atoms + [item]
        return LammpsPotentialFile(
            potential_df=potential_df,
            default_df=self._default_df,
            selected_atoms=selected_atoms,
            resource_path=self._resource_path,
        )


class PotentialAvailable:
    def __init__(self, list_of_potentials):
        self._list_of_potentials = {
            "pot_" + v.replace("-", "_").replace(".", "_"): v
            for v in list_of_potentials
        }

    def __getattr__(self, name):
        if name in self._list_of_potentials:
            return self._list_of_potentials[name]
        else:
            raise AttributeError

    def __dir__(self):
        return list(self._list_of_potentials.keys())

    def __repr__(self):
        return str(dir(self))


def find_potential_file_base(path, resource_path_lst, rel_path):
    if path is not None:
        for resource_path in resource_path_lst:
            path_direct = os.path.join(resource_path, path)
            path_indirect = os.path.join(resource_path, rel_path, path)
            if os.path.exists(path_direct):
                return path_direct
            elif os.path.exists(path_indirect):
                return path_indirect
    raise ValueError(
        "Either the filename or the functional has to be defined.",
        path,
        resource_path_lst,
    )


def view_potentials(structure: Atoms, resource_path: str) -> pandas.DataFrame:
    """
    List all interatomic potentials for the given atomistic structure including all potential parameters.

    To quickly get only the names of the potentials you can use `list_potentials()` instead.

    Args:
        structure (Atoms): The structure for which to get potentials.
        resource_path (str): Path to the "lammps/potentials_lammps.csv" file

    Returns:
        pandas.Dataframe: Dataframe including all potential parameters.
    """
    list_of_elements = set(structure.get_chemical_symbols())
    return LammpsPotentialFile(resource_path=resource_path).find(list_of_elements)


def convert_path_to_abs_posix(path: str) -> str:
    """
    Convert path to an absolute POSIX path

    Args:
        path (str): input path.

    Returns:
        str: absolute path in POSIX format
    """
    return (
        Path(path.strip())
        .expanduser()
        .resolve()
        .absolute()
        .as_posix()
        .replace("\\", "/")
    )


def update_potential_paths(
    df_pot: pandas.DataFrame, resource_path: str
) -> pandas.DataFrame:
    config_lst = []
    for row in df_pot.itertuples():
        potential_file_lst = row.Filename
        potential_file_path_lst = [
            os.path.join(resource_path, f) for f in potential_file_lst
        ]
        potential_dict = {os.path.basename(f): f for f in potential_file_path_lst}
        potential_commands = []
        for line in row.Config:
            line = line.replace("\n", "")
            for key, value in potential_dict.items():
                line = line.replace(key, value)
            potential_commands.append(line)
        config_lst.append(potential_commands)
    df_pot["Config"] = config_lst
    return df_pot


def get_resource_path_from_conda(
    env_variables: tuple[str] = ("CONDA_PREFIX", "CONDA_DIR"),
) -> str:
    env = os.environ
    for conda_var in env_variables:
        if conda_var in env:
            resource_path = os.path.join(env[conda_var], "share", "iprpy")
            if os.path.exists(resource_path):
                return resource_path
    raise ValueError("No resource_path found" + potential_installation)


def get_potential_dataframe(structure: Atoms, resource_path=None):
    if resource_path is None:
        resource_path = get_resource_path_from_conda()
    return update_potential_paths(
        df_pot=view_potentials(structure=structure, resource_path=resource_path),
        resource_path=resource_path,
    )


def get_potential_by_name(potential_name: str, resource_path: Optional[str] = None):
    if resource_path is None:
        resource_path = get_resource_path_from_conda()
    df = LammpsPotentialFile(resource_path=resource_path).list()
    return update_potential_paths(
        df_pot=df[df.Name == potential_name], resource_path=resource_path
    ).iloc[0]


def validate_potential_dataframe(
    potential_dataframe: pandas.DataFrame,
) -> pandas.DataFrame:
    if isinstance(potential_dataframe, pandas.Series):
        return potential_dataframe
    elif isinstance(potential_dataframe, pandas.DataFrame):
        if len(potential_dataframe) == 1:
            return potential_dataframe.iloc[0]
        elif len(potential_dataframe) == 0:
            raise ValueError(
                "The potential_dataframe is an empty pandas.DataFrame:",
                potential_dataframe,
            )
        else:
            raise ValueError(
                "The potential_dataframe contains more than one interatomic potential, please select one:",
                potential_dataframe,
            )
    else:
        raise TypeError(
            "The potential_dataframe should be a pandas.DataFrame or pandas.Series, but instead it is of type:",
            type(potential_dataframe),
        )



import pyiron_lammps._version
from pyiron_lammps.compatibility.calculate import calc_md, calc_minimize, calc_static
from pyiron_lammps.compatibility.file import (
    lammps_file_initialization,
    lammps_file_interface_function,
)
from pyiron_lammps.output import parse_lammps_output as parse_lammps_output_files
from pyiron_lammps.potential import (
    get_potential_by_name,
    get_potential_dataframe,
    validate_potential_dataframe,
)
from pyiron_lammps.structure import write_lammps_datafile as write_lammps_structure

DUMP_COMMANDS = [
    "dump 1 all custom 100 dump.out id type xsu ysu zsu fx fy fz vx vy vz\n",
    'dump_modify 1 sort id format line "%d %d %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g"\n',
]

THERMO_COMMANDS = [
    "thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol\n",
    "thermo_modify format float %20.15g\n",
    "thermo 100\n",
]

__version__ = pyiron_lammps._version.__version__
__all__ = [
    "calc_md",
    "calc_minimize",
    "calc_static",
    "get_potential_by_name",
    "get_potential_dataframe",
    "lammps_file_initialization",
    "lammps_file_interface_function",
    "parse_lammps_output_files",
    "validate_potential_dataframe",
    "write_lammps_structure",
]
